# This python file containts the class objects related to the
# tree representation of inter-temporal links between problems

# Initial developer of this file Dr. Jose Nicolas Melchor Gutierrez


import networkx as nx


class TemporalTree():
    
    G = nx.DiGraph()